/*-
 * Copyright (c) 2008 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Alistair Crooks (agc@NetBSD.org)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/time.h>

#include <netinet/in.h>

#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#ifdef USE_HISTEDIT
#  ifdef HAVE_HISTEDIT_H
#    include <histedit.h>
#  endif
#endif
#ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
#endif
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#ifdef HAVE_VIS_H
#  include <vis.h>
#endif

#include "libicb.h"

/*
 * Information for the protocol implementation here was taken from:
 *
 *	http://www.icb.net/_jrudd/icb/protocol.html
 *
 * and from interacting with the i2cbd server (which, interestingly
 * enough, disagrees with the protocol spec over such things as
 * "wh" command output.
 */

enum {
	M_LOGIN		= 'a',		/* login packet */
	M_LOGINOK	= 'a',		/* login packet */
	M_OPEN		= 'b',		/* open msg to group */
	M_PERSONAL	= 'c',		/* personal msg */
	M_STATUS	= 'd',		/* status update message */
	M_ERROR		= 'e',		/* error message */
	M_IMPORTANT	= 'f',		/* important announcement */
	M_EXIT		= 'g',		/* server exit */
	M_COMMAND	= 'h',		/* send a command from user */
	M_CMDOUT	= 'i',		/* output from a command */
	M_PROTO		= 'j',		/* protocol version information */
	M_BEEP		= 'k',		/* beeps */
	M_PING		= 'l',		/* ping packet */
	M_PONG		= 'm',		/* return for ping packet */
	M_NOOP		= 'n',		/* client no-op */

	SEP		= '\001',	/* protocol separator */
	ESC		= '\033'	/* Escape character */
};

enum {
	REASONABLE_LENGTH	= 64,
	MAX_ICB_MSG		= 256
};

/* an icb message */
typedef struct msg_t {
	int	 msgc;			/* # of message fields */
	char	*msgv[MAX_ICB_MSG];	/* the messages */
	uint8_t	 msglen[MAX_ICB_MSG];	/* length of the fields */
} msg_t;

#if 0
/* show all the characters in a message, even unprintables */
static void
show(char *src)
{
	char	dest[1024];

	strvis(dest, src, VIS_SAFE);
	printf("%s\n", dest);
}
#endif

/* save 'n' chars of 's' in allocated memory */
static char *
strnsave(char *s, int n)
{
	char	*cp;

	if (n < 0) {
		n = strlen(s);
	}
	if ((cp = calloc(1, n + 1)) != NULL) {
		(void) memcpy(cp, s, n);
		cp[n] = 0x0;
	}
	return cp;
}

/* split a message into constituent parts */
static void
splitmsg(char *s, msg_t *msg)
{
	uint8_t	 cc;
	uint8_t	 len;
	int	 i;

	(void) memset(msg, 0x0, sizeof(*msg));
	cc = *s;
	msg->msgv[0] = &s[2];
	for (len = 0, i = 2 ; i < cc ; i++) {
		switch(s[i]) {
		case SEP:
			msg->msglen[msg->msgc++] = len;
			msg->msgv[msg->msgc] = &s[i + 1];
			len = 0;
			break;
		default:
			len += 1;
			break;
		}
	}
	msg->msglen[msg->msgc++] = len;
}

/* connect the socket to the server */
static int 
sock_connect(int sock, const char *hostname, const int port)
{
	struct addrinfo	 hints;
	struct addrinfo	*res;
	char		 portstr[32];
	int              rc = 0;

	(void) memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_INET;
	hints.ai_socktype = SOCK_STREAM;
	(void) snprintf(portstr, sizeof(portstr), "%d", port);
	/* Attempt connection */
#ifdef AI_NUMERICSERV
	hints.ai_flags = AI_NUMERICSERV;
#endif
	if ((rc = getaddrinfo(hostname, portstr, &hints, &res)) != 0) {
		hints.ai_flags = 0;
		if ((rc = getaddrinfo(hostname, "icbd", &hints, &res)) != 0) {
			(void) fprintf(stderr, "getaddrinfo: %s", gai_strerror(rc));
			return 0;
		}
	}
	if ((rc = connect(sock, res->ai_addr, res->ai_addrlen)) < 0) {
		(void) fprintf(stderr, "connection failed\n");
		freeaddrinfo(res);
		return -1;
	}
	freeaddrinfo(res);
	if (rc < 0) {
		(void) fprintf(stderr, "connect() to %s:%d failed (rc %d)\n", hostname, port, rc);
	}
	return rc;
}

/* hash function */
static uint32_t
djbhash(char *str, int len)
{
	uint32_t	hash = 5381;
	uint32_t	i;

	for (i = 0; i < len; str++, i++) {
		hash = ((hash << 5) + hash) + (*str);
	}
	return hash;
}

/* print the timestamp */
static char *
fmttime(icbchan_t *chan)
{
	static char	 buf[32];
	time_t		 t;
	char		*cp;

	if (chan->timestamps) {
		(void) time(&t);
		cp = ctime(&t);
		if (chan->verbose) {
			(void) snprintf(buf, sizeof(buf), "[%.19s] ", cp);
		} else {
			(void) snprintf(buf, sizeof(buf), "[%.3s %.8s] ", cp, &cp[11]);
		}
	} else {
		buf[0] = 0x0;
	}
	return buf;
}

/* find the natural length for output */
static int
getlen(const char *text, int off, int cc, int maxlen)
{
	int	len;
	int	i;

	if (cc <= off + maxlen) {
		/* if we just have less than line length, use it */
		return cc - off;
	}
	/* do an oscillating search for whitespace */
	for (len = off + maxlen, i = 0 ; i < maxlen ; i++) {
		if (text[len + i] == ' ') {
			return len + i + 1 - off;
		}
		if (text[len - i] == ' ') {
			return len - i + 1 - off;
		}
	}
	/* force cut */
	return maxlen;
}

/* check for any searchable triggers */
static void
showtext(icbchan_t *chan, char *name, char *s, int len)
{
	regmatch_t	 matchv[10];
	char		*buf;
	char		*cp;
	int		 linelen;
	int		 done;
	int		 off;
	int		 cc;
	int		 i;
	int		 j;

	cc = 0;
	buf = cp = malloc(MAX_ICB_MSG * 4);
	for (done = i = 0 ; !done && i < chan->triggers.c ; i++) {
		/* look for first trigger */
		if (regexec(&chan->triggers.v[i].r, s, 10, matchv, 0) == 0) {
			for (j = 0 ; j < matchv[0].rm_so ; j++) {
				*cp++ = s[j];
			}
			cp += snprintf(cp, (MAX_ICB_MSG * 4) - (int)(cp - buf), "%c[3m", ESC);
			for ( ; j < matchv[0].rm_eo ; j++) {
				*cp++ = s[j];
			}
			cp += snprintf(cp, (MAX_ICB_MSG * 4) - (int)(cp - buf), "%c[m", ESC);
			for ( ; j < len ; j++) {
				*cp++ = s[j];
			}
			*cp++ = 0x0;
			/* only embolden the first trigger */
			/* that will draw attention */
			/* no need to make output look like Christmas lights */
			cc = (int)(cp - buf);
			done = 1;
		}
	}
	if (!done) {
		(void) strlcpy(buf, s, MAX_ICB_MSG * 4);
		cc = len;
	}
	/* printf("%s%s\n", name, buf); */
	for (off = 0 ; off < cc ; off += linelen) {
		linelen = getlen(buf, off, cc, REASONABLE_LENGTH);
		printf("%s%.*s\n", name, linelen, &buf[off]);
	}
	(void) free(buf);
}

/* show a personal or an open message */
static void
showmsg(icbchan_t *chan, char *nick, char type)
{
	regmatch_t	 matchv[10];
	const char	*personal;
	char		 name[256];
	char		*sep;
	char		 beg[12];
	char		 end[12];
	int		 hash;
	int		 i;

	if ((sep = strchr(nick, SEP)) != NULL) {
		if (type == M_OPEN) {
			for (i = 0 ; i < chan->filters.c ; i++) {
				if (regexec(&chan->filters.v[i].r, sep + 1, 10, matchv, 0) == 0) {
					if (chan->verbose) {
						printf("[ignoring \"%s\" via \"%s\" filter]\n",
							sep + 1, chan->filters.v[i].text);
					}
					return;
				}
			}
		}
		if (chan->colours) {
			/* ANSI colors range from 30 to 37, but 30 is black */
			hash = (djbhash(nick, (int)(sep - nick)) % 7) + 31;
			(void) snprintf(beg, sizeof(beg), "%c[%dm", ESC, hash);
			(void) snprintf(end, sizeof(end), "%c[m", ESC);
		} else {
			beg[0] = end[0] = 0x0;
		}
		personal = (type == M_PERSONAL) ? "*" : "";
		(void) snprintf(name, sizeof(name), "%s%s<%s%.*s%s>%s ",
			fmttime(chan),
			beg, personal, (int)(sep - nick), nick, personal, end);
		showtext(chan, name, sep + 1, strlen(sep + 1));
	}
}

/* show command output */
static void
showcmdout(icbchan_t *chan, char *s)
{
	time_t	t;
	char	tim[20];
	msg_t	msg;
	int	idle;
	int	i;

	splitmsg(s, &msg);
	if (strncmp(msg.msgv[0], "co", 2) == 0) {
		for (i = 1 ; i < msg.msgc ; i++) {
			printf("%s%.*s\n", fmttime(chan), (int)msg.msglen[i], msg.msgv[i]);
		}
	} else if (strncmp(msg.msgv[0], "wl", 2) == 0) {
		idle = atoi(msg.msgv[3]);
		t = atoi(msg.msgv[5]);
		(void) memcpy(tim, ctime(&t), 19);
		tim[19] = 0x0;
		printf("%s%c%12.*s   %2dh%02dm%02ds  %.19s  %.*s@%.*s %.*s\n",
			fmttime(chan),
			*msg.msgv[1],
			msg.msglen[2], msg.msgv[2],
			idle / (60 * 60), (idle / 60) % 60, idle % 60,
			tim,
			msg.msglen[6], msg.msgv[6],
			msg.msglen[7], msg.msgv[7],
			msg.msglen[8], msg.msgv[8]);
	} else if (strncmp(msg.msgv[0], "wh", 2) == 0) {
		printf("%s     Nickname        Idle  Login                Identity\n", fmttime(chan));
	} else {
		if (chan->verbose) {
			printf("%sweird %2.2s\n", fmttime(chan), msg.msgv[0]);
		}
	}
}

/* send the equivalent pong message back */
static void
sendpong(icbchan_t *chan, char *buf)
{
	int	cc;

	buf[1] = M_PONG;
	if ((cc = send(chan->sock, buf, buf[0], 0)) != buf[0]) {
		warn("short send - expected %d chars, got %d\n", buf[0], cc);
	}
}

/* send a command to the server */
static void
sendcommand(icbchan_t *chan, char *s)
{
	char	buf[256];
	char	*cp;
	int	got;
	int	cc;

	cc = snprintf(&buf[1], sizeof(buf) - 1, "%c%s", M_COMMAND, s) + 1;
	/* include terminating NUL */
	buf[0] = (uint8_t)cc;
	if ((cp = strchr(&buf[2], ' ')) != NULL) {
		*cp = SEP;
	}
	if ((got = send(chan->sock, buf, cc + 1, 0)) != cc + 1) {
		warn("short send - expected %d chars, got %d\n", cc + 1, got);
	}
}

/* send a message to the server */
static void
sendpermanent(icbchan_t *chan, char *s)
{
	char	buf[256];
	int	got;
	int	cc;

	cc = snprintf(&buf[1], sizeof(buf) - 1, "%cm%c%s %s", M_COMMAND, SEP, chan->permanent, s) + 1;
	/* include terminating NUL */
	buf[0] = (uint8_t)cc;
	if ((got = send(chan->sock, buf, cc + 1, 0)) != cc + 1) {
		warn("short send - expected %d chars, got %d\n", cc + 1, got);
	}
}

/* send the contents of a line */
static int
sendopenline(icbchan_t *chan, const char *text, int off, int len)
{
	char	buf[BUFSIZ];
	int	got;
	int	cc;

	cc = snprintf(&buf[1], sizeof(buf) - 1, "%c%.*s", M_OPEN, len, &text[off]) + 1;
	/* include terminating NUL */
	buf[0] = (uint8_t)(cc);
	if ((got = send(chan->sock, buf, cc + 1, 0)) != cc + 1) {
		warn("short send - expected %d chars, got %d\n", cc + 1, got);
		return 0;
	}
	return 1;
}

/* set the permanent recipient of messages */
static void
setpermanent(icbchan_t *chan, char *nick)
{
	if (nick == NULL) {
		if (chan->permanent) {
			(void) free(chan->permanent);
			chan->permanent = NULL;
		}
	} else {
		while (*nick && isspace((unsigned char)*nick)) {
			nick++;
		}
		chan->permanent = strdup(nick);
	}
}

/* add a pattern to embolden or ignore */
static int
addpattern(icbarray_t *arr, char *pat, int delimited, const char *type)
{
	icbpat_t	*tp;
	char		*end;
	char		 sep;
	int		 newsize;

	if (arr->size == 0) {
		arr->size = 15;
		if ((arr->v = calloc(15, sizeof(icbpat_t))) == NULL) {
			return 0;
		}
	} else if (arr->c == arr->size) {
		newsize = arr->size * 2;
		if ((tp = realloc(arr->v, newsize * sizeof(icbpat_t))) == NULL) {
			return 0;
		}
		arr->size = newsize;
		arr->v = tp;
	}
	for ( ; *pat && isspace((unsigned char)*pat) ; pat++) {
	}
	if (delimited) {
		for (sep = *pat++, end = pat ; *end && *end != sep ; end++) {
			if (*end == '\\') {
				end += 1;
			}
		}
		arr->v[arr->c].text = strnsave(pat, (int)(end - pat));
	} else {
		arr->v[arr->c].text = strnsave(pat, -1);
	}
	(void) regcomp(&arr->v[arr->c].r, arr->v[arr->c].text, REG_EXTENDED);
	if (type != NULL) {
		printf("[Added %s pattern %d for \"%s\"]\n", type, arr->c, arr->v[arr->c].text);
	}
	arr->c += 1;
	return 1;
}

/* remove a pattern */
static int
delpattern(icbarray_t *arr, char *pat, const char *type)
{
	int	i;

	for ( ; *pat && isspace((unsigned char)*pat) ; pat++) {
	}
	for (i = 0 ; i < arr->c ; i++) {
		if (strcmp(arr->v[i].text, pat) == 0) {
			printf("[Removing %s pattern %d for \"%s\"]\n", type, i, pat);
			(void) free(arr->v[i].text);
			arr->c -= 1;
			if (i < arr->c) {
				(void) memmove(&arr->v[i], &arr->v[i + 1], arr->c * sizeof(icbpat_t));
			}
			return 1;
		}
	}
	return 0;
}

/* add a nickname to the list of presonal nicks */
static void
addpersonal(icbchan_t *chan, char *nick)
{
	char	*cp;

	while (*nick && isspace((unsigned char)*nick)) {
		nick++;
	}
	for (cp = nick + 1 ; *cp && !isspace((unsigned char)*cp) ; cp++) {
	}
	if (chan->prevpersonal) {
		(void) free(chan->prevpersonal);
	}
	chan->prevpersonal = strnsave(nick, (int)(cp - nick));
}

#ifdef USE_HISTEDIT
static char	prompt[48];

/* set the prompt */
static void
setprompt(icbchan_t *chan)
{
	(void) chan;
#if 0
	if (chan->permanent) {
		(void) snprintf(prompt, sizeof(prompt), "%s@%s#%s <*%s*> ",
			chan->nick, chan->server, chan->group, chan->permanent);
	} else {
		(void) snprintf(prompt, sizeof(prompt), "%s@%s#%s ",
			chan->nick, chan->server, chan->group);
	}
#else
	(void) strlcpy(prompt, "libicb> ", sizeof(prompt));
#endif
}

/* print the prompt */
static char *
getprompt(EditLine *el)
{
	(void) el;
	return prompt;
}
#endif

/* open a channel to the icb server */
int
icb_chanopen(icbchan_t *chan, icbchan_t *defaults)
{
	int	on = 1;

	(void) memset(chan, 0x0, sizeof(*chan));
	/* copy necessary fields from defaults to chan */
	chan->server = strdup(defaults->server);
	chan->group = strdup(defaults->group);
	chan->id = strdup(defaults->id);
	chan->nick = strdup(defaults->nick);
	chan->passwd = strdup(defaults->passwd);
	chan->verbose = defaults->verbose;
	chan->colours = defaults->colours;
	chan->timestamps = defaults->timestamps;
	chan->cmdchar = defaults->cmdchar;
	/* set up basic nickname trigger */
	addpattern(&chan->triggers, chan->nick, 0, "trigger");
	/* tell what we're doing */
	(void) fprintf(stderr, "Connecting to port %d of host %s.\n", defaults->port, defaults->server);
	/* create a socket */
	if ((chan->sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		(void) fprintf(stderr,"libicb: can't get a socket\n");
		return 0;
	}
	if (setsockopt(chan->sock, SOL_SOCKET,
		SO_KEEPALIVE, (char *)&on, sizeof(on)) < 0) {
		(void) fprintf(stderr,"libicb: can't set KEEPALIVE\n");
		return 0;
	}
	if (setsockopt(chan->sock, SOL_SOCKET,
		SO_REUSEADDR, (char *)&on, sizeof(on)) < 0) {
		(void) fprintf(stderr,"libicb: can't set REUSEADDR\n");
		return 0;
	}
	/* connect to the icbd server */
	if (sock_connect(chan->sock, defaults->server, defaults->port) != 0) {
		(void) fprintf(stderr,"libicb: couldn't connect to %s server.\n", defaults->server);
		return 0;
	}
	return 1;
}

/* login to an icb server */
int
icb_login(icbchan_t *chan)
{
	char	buf[BUFSIZ];
	int	got;
	int	cc;

	cc = snprintf(&buf[1], sizeof(buf) - 1, "%c%s%c%s%c%s%c%s%c%s",
		M_LOGIN, chan->id, SEP,
		chan->nick, SEP,
		chan->group, SEP,
		"login", SEP,
		chan->passwd) + 1;
	/* include terminating NUL */
	buf[0] = (uint8_t)cc;
	if ((got = send(chan->sock, buf, cc + 1, 0)) != cc + 1) {
		warn("short send - expected %d chars, gotr %d\n", cc + 1, got);
		return 0;
	}
	return 1;
}

/* send open message text to server */
int
icb_sendopenmsg(icbchan_t *chan, const char *text)
{
	int	off;
	int	len;
	int	cc;

	/* take any newline character at the end of text off */
	cc = strlen(text);
	if (text[cc - 1] == '\n') {
		cc -= 1;
	}
	for (off = 0 ; off < cc ; off += len) {
		/* find a reasonable place to break the line */
		len = getlen(text, off, cc, MAX_ICB_MSG - 6);
		if (!sendopenline(chan, text, off, len)) {
			return 0;
		}
	}
	return 1;
}

/* this structure describes a command alias */
typedef struct alias_t {
	const char	*clientcmd;	/* client command */
	const int	 clientlen;	/* its length */
	const char	*servcmd;	/* what the server will recognise */
	const int	 servlen;	/* its length */
} alias_t;

static alias_t	aliases[] = {
	{	"nick",	4,	"name",	4	},
	{	NULL,	0,	NULL,	0	}
};

/* make sure we handle alias commands */
static int
expandcmd(char *buf, int cc)
{
	alias_t	*alias;
	char	 was[BUFSIZ];

	(void) strlcpy(was, buf, sizeof(was));
	for (alias = aliases ; alias->clientcmd ; alias++) {
		if (strncmp(buf, alias->clientcmd, alias->clientlen) == 0) {
			(void) memcpy(buf, alias->servcmd, alias->servlen);
			(void) strlcpy(&buf[alias->servlen], &was[alias->clientlen],
				sizeof(buf) - alias->servlen);
			return 1;
		}
	}
	return 0;
}

/* read the file, and then send the contents to the server at the desired speed */
static int
readfile(icbchan_t *chan, char *fname)
{
	struct stat	 st;
	uint64_t	 seen;
	size_t		 cc;
	char	 	 buf[1024];
	FILE		*fp;
	int		 s;

	if ((fp = fopen(fname, "r")) == NULL) {
		(void) fprintf(stderr, "File \"%s\" does not exist\n", fname);
		return 0;
	}
	printf("[PID %d: Reading file \"%s\"]\n", getpid(), fname);
	(void) fstat(fileno(fp), &st);
	for (seen = 0 ; fgets(buf, sizeof(buf), fp) != NULL ; seen += cc) {
		icb_sendopenmsg(chan, buf);
		cc = strlen(buf);
		s = cc / icb_getspeed();
		printf("%.*s [%" PRIu64 "/%" PRIu64 ", %" PRIu64 "%%, %d secs]\n",
			(int)(cc - 1), buf, seen, st.st_size,
			(seen * 100) / st.st_size, s);
		while ((s = sleep(s)) > 0) {
		}

	}
	(void) fclose(fp);
	return 1;
}

/* have a conversation via icb */
int
icb_chat(icbchan_t *chan)
{
#ifdef USE_TERMIOS
	struct termios	 savedtty;
	struct termios	 tty;
#endif
	struct timeval	 tv;
	unsigned char	 buf[BUFSIZ];
	unsigned char	*cp;
#ifdef USE_HISTEDIT
	const char	*line;
	HistEvent 	 ev;
	EditLine	*el;
	History		*hist;
#endif
	fd_set		 reads;
	msg_t		 msg;
	unsigned	 want;
	int		 done;
	int		 got;
	int		 cc;

	cp = NULL;
#ifdef USE_HISTEDIT
	/* init editline */
	setprompt(chan);
	el = el_init(getprogname(), stdin, stdout, stderr);
	hist = history_init();			/* init the builtin history */
	history(hist, &ev, H_SETSIZE, 1000);	/* remember 1000 events */
	el_set(el, EL_HIST, history, hist);	/* use history */
	el_set(el, EL_EDITOR, "emacs");		/* default editor is emacs */
	el_set(el, EL_EDITMODE, 1);		/* say that we are editing */
	el_set(el, EL_SIGNAL, 1);		/* use libedit's signal handlers */
	el_set(el, EL_PROMPT, &getprompt);	/* set the prompt functions */
	el_source(el, NULL);			/* read ~/.editrc */
	el_reset(el);				/* let's reset at the start */
#endif
#ifdef USE_TERMIOS
	tcgetattr(STDIN_FILENO, &savedtty);
	(void) memcpy(&tty, &savedtty, sizeof(tty));
	tty.c_lflag &= ~(ECHO | ECHOK | ICANON);
	tty.c_cc[VTIME] = 1;
	tcsetattr(STDIN_FILENO, TCSANOW, &tty);
#endif
	want = 0;
	got = 0;
	for (done = 0 ; !done ; ) {
		FD_ZERO(&reads);
		FD_SET(STDIN_FILENO, &reads);
		FD_SET(chan->sock, &reads);
		tv.tv_sec = 30000;
		tv.tv_usec = 0;
		tcsetattr(STDIN_FILENO, TCSANOW, &tty);
		if (select(32, &reads, NULL, NULL, &tv) == 0) {
			continue;
		}
		if (FD_ISSET(STDIN_FILENO, &reads)) {
			/* user input */
#ifdef USE_HISTEDIT
			cc = 0;
			line = el_gets(el, &cc);
			if (cc == 0 || line == NULL) {
				/* got eof */
				done = 1;
				continue;
			}
			(void) memcpy(buf, line, cc);
			if (buf[--cc] == '\n') {
				buf[cc] = 0x0;
			}
			history(hist, &ev, H_ENTER, buf);
#else
			if (fgets(buf, sizeof(buf), stdin) == NULL) {
				done = 1;
				continue;
			}
			cc = strlen(buf) - 1;
			buf[cc] = 0x0;
#endif
			if (buf[0] == chan->cmdchar) {
				switch(buf[1]) {
				case 'C':
					/* clear a permanent recipient */
					printf("[Clearing permanent messaging]\n");
					setpermanent(chan, NULL);
					break;
				case 'M':
					/* set a permanent recipient for personal messages */
					setpermanent(chan, (char *)&buf[2]);
					addpersonal(chan, (char *)&buf[2]);
					printf("[Permanent messaging to %s]\n", &buf[3]);
					break;
				case 'S':
					/* say a message while permanent is in effect */
					icb_sendopenmsg(chan, (char *)&buf[2]);
					break;
				case 'F':
					delpattern(&chan->filters, (char *)&buf[2], "filter");
					break;
				case 'f':
					addpattern(&chan->filters, (char *)&buf[2], 1, "filter");
					break;
				case 'm':
					/* message someone */
					sendcommand(chan, (char *)&buf[1]);
					addpersonal(chan, (char *)&buf[2]);
					break;
				case 'q':
					/* quit */
					if (cc == 2) {
						done = 1;
						continue;
					}
					break;
				case 'r':
					/* do the readfile bit from speak */
					readfile(chan, (char *)&buf[2]);
					break;
				case 'T':
					delpattern(&chan->triggers, (char *)&buf[2], "trigger");
					break;
				case 't':
					addpattern(&chan->triggers, (char *)&buf[2], 1, "trigger");
					break;
				default:
					expandcmd((char *)&buf[1], cc);
					sendcommand(chan, (char *)&buf[1]);
					break;
				}
			} else if (buf[0] == '\t') {
				char	newbuf[MAX_ICB_MSG * 4];

				if (chan->prevpersonal == NULL) {
					printf("[No previous personal recipient]\n");
				} else {
					(void) snprintf(newbuf, sizeof(newbuf), "m %s %s",
						chan->prevpersonal, &buf[1]);
					sendcommand(chan, newbuf);
				}
			} else {
				if (chan->permanent) {
					sendpermanent(chan, (char *)buf);
				} else {
					icb_sendopenmsg(chan, (char *)buf);
				}
			}
		}
		if (FD_ISSET(chan->sock, &reads)) {
			/* server input */
			if (want == 0) {
				if ((cc = read(chan->sock, buf, 1)) < 0) {
					continue;
				}
				want = (unsigned)(buf[0]);
				cp = &buf[1];
				got = 0;
			}
			if ((cc = read(chan->sock, &cp[got], want - got)) < 0) {
				(void) fprintf(stderr, "Bad read\n");
				done = 1;
				continue;
			}
			got += cc;
			if (got < want) {
				continue;
			}
			/* set this to 0 for next time around the loop */
			want = 0;
			switch(buf[1]) {
			case M_LOGINOK:
				printf("%s[Received login OK from server]\n", fmttime(chan));
				break;
			case M_OPEN:
			case M_PERSONAL:
				showmsg(chan, (char *)&buf[2], buf[1]);
				break;
			case M_PROTO:
				splitmsg((char *)buf, &msg);
				printf("Connected to the %.*s ICB server %.*s (protocol %c)\n",
					(int)msg.msglen[1], msg.msgv[1],
					(int)msg.msglen[2], msg.msgv[2],
					 buf[2]);
				break;
			case M_ERROR:
				printf("%s[=Error=] %s\n", fmttime(chan), &buf[2]);
				break;
			case M_BEEP:
				printf("%s\a[=Beep!=] %s sent you a beep.\n", fmttime(chan), &buf[2]);
				break;
			case M_PING:
				printf("%s[Received ping]\n", fmttime(chan));
				sendpong(chan, (char *)buf);
				break;
			case M_CMDOUT:
				showcmdout(chan, (char *)buf);
				break;
			case M_IMPORTANT:
			case M_STATUS:
				splitmsg((char *)buf, &msg);
				printf("%s%s[=%s%.*s=] %.*s\n",
					fmttime(chan),
					(buf[1] == M_IMPORTANT) ? "\a" : "",
					(buf[1] == M_IMPORTANT) ? "Important=" : "",
					(int)msg.msglen[0], msg.msgv[0],
					(int)msg.msglen[1], msg.msgv[1]);
				break;
			case M_EXIT:
				printf("%s[Received EXIT from server]\n", fmttime(chan));
				done = 1;
				continue;
			default:
				printf("%s[Received unusual message from server (%d bytes) '%c']\n",
					fmttime(chan), buf[0], buf[1]);
				showmsg(chan, (char *)&buf[1], M_OPEN);
				break;
			}
		}
	}
#ifdef USE_HISTEDIT
	el_end(el);
#endif
#ifdef USE_TERMIOS
	tcsetattr(STDIN_FILENO, TCSANOW, &savedtty);
#endif
	return 1;
}

/* read commands from an rcfile */
int
icb_rcfile(icbchan_t *chan, const char *f)
{
	FILE	*fp;
	char	 buf[BUFSIZ];
	char	*cp;

	if ((fp = fopen(f, "r")) == NULL) {
		return 0;
	}
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		buf[strlen(buf) - 1] = 0x0;
		for (cp = buf ; *cp && isspace((unsigned char)*cp) ; cp++) {
		}
		if (*cp == 0x0 || *cp == '#') {
			continue;
		}
		if (*cp == chan->cmdchar) {
			if (strncmp(cp + 1, "cmdchar", 7) == 0) {
				for (cp += 1 + 7 ; *cp && isspace((unsigned char)*cp) ; cp++) {
				}
				if (*cp) {
					chan->cmdchar = *cp;
				}
				printf("[Command character is now %c]\n", chan->cmdchar);
			} else if (strncmp(cp + 1, "echoback", 8) == 0) {
				sendcommand(chan, cp + 1);
				printf("[Echoback is now %s]\n", cp + 1 + 8);
			} else if (strncmp(cp + 1, "trigger", 7) == 0) {
				addpattern(&chan->triggers, cp + 7 + 1, 1, "trigger");
			} else if (strncmp(cp + 1, "untrigger", 9) == 0) {
				delpattern(&chan->triggers, cp + 9 + 1, "trigger");
			} else if (strncmp(cp + 1, "filter", 6) == 0) {
				addpattern(&chan->filters, cp + 6 + 1, 1, "filter");
			} else if (strncmp(cp + 1, "unfilter", 8) == 0) {
				delpattern(&chan->filters, cp + 8 + 1, "filter");
			}
		}
	}
	(void) fclose(fp);
	return 1;
}

static int	speed;

void
icb_setspeed(const int n)
{
	printf("Moving from speed %d to %d\n", speed, n);
	speed = n;
}

int
icb_getspeed(void)
{
	return speed;
}
