/*-
 * Copyright (c) 2008 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Alistair Crooks (agc@NetBSD.org)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "libicb.h"

/* this struct describes an entity to be monitored */
typedef struct ent_t {
	char	prevname[MAXPATHLEN];	/* old file name */
	char	newname[MAXPATHLEN];	/* new file name */
	char	feed[MAXPATHLEN];	/* RSS Feed as a URL */
} ent_t;

static ent_t	entv[32];
static int	entc;

/* initialise the monitoring entity */
static int
initent(ent_t *ep, const char *f)
{
	char	ch;
	int	fd;

	(void) memset(ep, 0x0, sizeof(*ep));
	(void) strlcpy(ep->feed, f, sizeof(ep->feed));
	(void) strlcpy(ep->prevname, "/tmp/rssbot.XXXXXX", sizeof(ep->prevname));
	fd = mkstemp(ep->prevname);
	ch = ' ';
	(void) write(fd, &ch, 1);
	(void) close(fd);
	return 1;
}

/* note the additions to the rss feed */
static int
diffent(icbchan_t *chan, ent_t *ep)
{
	FILE	*pp;
	char	 cmd[MAXPATHLEN * 3];
	char	 buf[BUFSIZ];
	int	 fd;

	(void) strlcpy(ep->newname, "/tmp/rssbot.XXXXXX", sizeof(ep->newname));
	fd = mkstemp(ep->newname);
	(void) snprintf(cmd, sizeof(cmd), "rss2html -o txt -d %s > %s", ep->feed, ep->newname);
	(void) system(cmd);
	(void) snprintf(cmd, sizeof(cmd), "diff %s %s", ep->prevname, ep->newname);
	if ((pp = popen(cmd, "r")) == NULL) {
		(void) fprintf(stderr, "can't run command \"%s\"\n", cmd);
	} else {
		while (fgets(buf, sizeof(buf), pp) != NULL) {
			if (buf[0] == '>') {
				icb_sendopenmsg(chan, &buf[2]);
			}
		}
		(void) pclose(pp);
	}
	(void) close(fd);
	(void) unlink(ep->prevname);
	/* make the newname the prevname */
	(void) strlcpy(ep->prevname, ep->newname, sizeof(ep->prevname));
	return 1;
}

int
main(int argc, char **argv)
{
	icbchan_t	 defaults;
	icbchan_t	 chan;
	char		*home;
	char		 buf[MAXPATHLEN];
	int		 sleepsecs;
	int		 i;

	(void) memset(&defaults, 0x0, sizeof(defaults));
	defaults.timestamps = 1;
	defaults.colours = 1;
	defaults.id = strdup("rssbot");
	defaults.nick = strdup("rssbot");
	defaults.port = 7326;
	defaults.server = strdup("localhost");
	defaults.group = strdup("monitor");
	defaults.passwd = strdup("");
	defaults.cmdchar = '/';
	home = getenv("HOME");
	sleepsecs = 1;
	while ((i = getopt(argc, argv, "g:h:n:p:t:v~:")) != -1) {
		switch(i) {
		case 'g':
			defaults.group = optarg;
			break;
		case 'h':
			defaults.server = optarg;
			break;
		case 'n':
			defaults.nick = optarg;
			break;
		case 'p':
			defaults.port = atoi(optarg);
			break;
		case 't':
			sleepsecs = atoi(optarg);
			break;
		case 'v':
			defaults.verbose += 1;
			break;
		case '~':
			home = optarg;
			break;
		default:
			/* XXX */
			break;
		}
	}
	for (i = optind ; i < argc ; i++) {
		initent(&entv[entc++], argv[i]);
	}
	icb_chanopen(&chan, &defaults);
	icb_login(&chan);
	(void) snprintf(buf, sizeof(buf), "%s/.rssbotrc", home);
	icb_rcfile(&chan, buf);
	for (;;) {
		for (i = 0 ; i < entc ; i++) {
			diffent(&chan, &entv[i]);
		}
		sleep(sleepsecs);
	}
	exit(EXIT_SUCCESS);
}
