/*-
 * Copyright (c) 2008 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Alistair Crooks (agc@NetBSD.org)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "libicb.h"

/* this struct describes an entity to be monitored */
typedef struct ent_t {
	char		 fname[MAXPATHLEN];	/* file name */
	struct stat	 st;			/* file info */
	FILE		*fp;			/* open file pointer */
} ent_t;

static ent_t	entv[32];
static int	entc;

/* stat the file, and note any differences between old and new */
static int
statfile(icbchan_t *chan, ent_t *ep)
{
	struct stat	st;
	char		buf[128];

	if (stat(ep->fname, &st) != 0) {
		(void) fprintf(stderr, "File \"%s\" no longer exists\n", ep->fname);
		return 0;
	}
	if (st.st_ino != ep->st.st_ino) {
		(void) fprintf(stderr, "File \"%s\" has turned over\n", ep->fname);
		(void) fclose(ep->fp);
		ep->fp = fopen(ep->fname, "r");
	}
	if (st.st_mtime != ep->st.st_mtime) {
		(void) fseek(ep->fp, ep->st.st_size, SEEK_SET);
		while (fgets(buf, sizeof(buf), ep->fp) != NULL) {
			icb_sendopenmsg(chan, buf);
		}
	}
	(void) memcpy(&ep->st, &st, sizeof(st));
	return 1;
}

/* initialise the monitoring entity */
static int
initfile(const char *f, ent_t *ep)
{
	(void) memset(ep, 0x0, sizeof(*ep));
	(void) strlcpy(ep->fname, f, sizeof(ep->fname));
	(void) stat(ep->fname, &ep->st);
	ep->fp = fopen(f, "r");
	return 1;
}

int
main(int argc, char **argv)
{
	icbchan_t	 defaults;
	icbchan_t	 chan;
	char		*home;
	char		 buf[MAXPATHLEN];
	int		 sleepsecs;
	int		 i;

	(void) memset(&defaults, 0x0, sizeof(defaults));
	defaults.timestamps = 1;
	defaults.colours = 1;
	defaults.id = strdup("tailbot");
	gethostname(buf, sizeof(buf));
	defaults.nick = strdup(buf);
	defaults.port = 7326;
	defaults.server = strdup("localhost");
	defaults.group = strdup("monitor");
	defaults.passwd = strdup("");
	defaults.cmdchar = '/';
	home = getenv("HOME");
	sleepsecs = 1;
	while ((i = getopt(argc, argv, "g:h:n:p:t:v~:")) != -1) {
		switch(i) {
		case 'g':
			defaults.group = optarg;
			break;
		case 'h':
			defaults.server = optarg;
			break;
		case 'n':
			defaults.nick = optarg;
			break;
		case 'p':
			defaults.port = atoi(optarg);
			break;
		case 't':
			sleepsecs = atoi(optarg);
			break;
		case 'v':
			defaults.verbose += 1;
			break;
		case '~':
			home = optarg;
			break;
		default:
			/* XXX */
			break;
		}
	}
	for (i = optind ; i < argc ; i++) {
		initfile(argv[i], &entv[entc++]);
	}
	icb_chanopen(&chan, &defaults);
	icb_login(&chan);
	(void) snprintf(buf, sizeof(buf), "%s/.tailbotrc", home);
	icb_rcfile(&chan, buf);
	for (;;) {
		for (i = 0 ; i < entc ; i++) {
			statfile(&chan, &entv[i]);
		}
		sleep(sleepsecs);
	}
	exit(EXIT_SUCCESS);
}
